@@ -607,58 +607,79 @@ def test_mncontour(grad, cl):
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_contour(grad):
-    # FIXME: check the result
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
     m.migrad()
-    m.contour("x", "y")
+    x, y, v = m.contour("x", "y")
+    X, Y = np.meshgrid(x, y)
+    assert_allclose(func0(X, Y), v.T)
 
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_profile(grad):
-    # FIXME: check the result
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
     m.migrad()
-    m.profile("y")
+
+    y, v = m.profile("y", subtract_min=False)
+    assert_allclose(func0(m.values[0], y), v)
+
+    v2 = m.profile("y", subtract_min=True)[1]
+    assert np.min(v2) == 0
+    assert_allclose(v - np.min(v), v2)
 
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_mnprofile(grad):
-    # FIXME: check the result
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
     m.migrad()
-    if grad is None:
-        m.mnprofile("y", subtract_min=True)
+
     with pytest.raises(ValueError):
         m.mnprofile("foo")
 
+    y, v, _ = m.mnprofile("y", size=10, subtract_min=False)
+    m2 = Minuit(func0, grad=grad, x=1.0, y=2.0)
+    m2.fixed[1] = True
+    v2 = []
+    for yi in y:
+        m2.values = (m.values[0], yi)
+        m2.migrad()
+        v2.append(m2.fval)
 
-def test_mnprofile_subtract():
-    m = Minuit(func0, x=1.0, y=2.0)
-    m.migrad()
-    m.mnprofile("y", subtract_min=True)
+    assert_allclose(v, v2)
 
-
-def test_profile_subtract():
-    m = Minuit(func0, x=1.0, y=2.0)
-    m.migrad()
-    m.profile("y", subtract_min=True)
+    y, v3, _ = m.mnprofile("y", size=10, subtract_min=True)
+    assert np.min(v3) == 0
+    assert_allclose(v - np.min(v), v3)
 
 
 def test_contour_subtract():
     m = Minuit(func0, x=1.0, y=2.0)
     m.migrad()
-    m.contour("x", "y", subtract_min=True)
+    v = m.contour("x", "y", subtract_min=False)[2]
+    v2 = m.contour("x", "y", subtract_min=True)[2]
+    assert np.min(v2) == 0
+    assert_allclose(v - np.min(v), v2)
 
 
 def test_mncontour_no_fmin():
-    m = Minuit(lambda x, y: 0, x=0, y=0)
-    m.errordef = 1
-    with pytest.raises(ValueError):
-        m.mncontour("x", "y")
+    m = Minuit(func0, x=0, y=0)
+
+    with pytest.raises(RuntimeError):
+        m.mncontour("x", "y")  # fails, because this is not a minimum
+
+    # succeeds
+    m.values = (2, 5)
+    c = m.mncontour("x", "y", size=10)
+
+    # compute reference to compare with
+    m2 = Minuit(func0, x=0, y=0)
+    m2.migrad()
+    c2 = m.mncontour("x", "y", size=10)
+
+    assert_allclose(c, c2)
 
 
 def test_mncontour_with_fixed_var():
-    m = Minuit(lambda x, y: 0, x=0, y=0)
+    m = Minuit(func0, x=0, y=0)
     m.errordef = 1
     m.fixed["x"] = True
     m.migrad()
@@ -962,8 +983,8 @@ def test_params():
 
     # these are the initial param states
     expected = (
-        Param(0, "x", 1.0, 3.0, None, False, True, False, False, False, None, None),
-        Param(1, "y", 2.0, 4.0, None, False, False, True, False, True, None, 10),
+        Param(0, "x", 1.0, 3.0, None, False, True, None, None),
+        Param(1, "y", 2.0, 4.0, None, False, False, None, 10),
     )
     assert m.params == expected
 
@@ -1081,12 +1102,13 @@ def test_hesse_without_migrad():
     m.values["x"] = 1
     m.hesse()
     assert m.errors["x"] == approx((1.0 / 14.0) ** 0.5, abs=1e-4)
-    assert m.fmin is None
+    assert m.fmin
 
     m = Minuit(lambda x: 0, 0)
     m.errordef = 1
-    with pytest.raises(RuntimeError):
-        m.hesse()
+    m.hesse()
+    assert not m.accurate
+    assert m.fmin.hesse_failed
 
 
 def test_edm_goal():
@@ -1321,10 +1343,16 @@ def test_tolerance():
     m.migrad()
     assert m.valid
     edm = m.fmin.edm
+    m.tol = 0
     m.reset()
-    m.tol = 1e-6
     m.migrad()
     assert m.fmin.edm < edm
+    m.reset()
+    m.tol = None
+    assert m.tol == 0.1
+    m.reset()
+    m.migrad()
+    assert m.fmin.edm == edm
 
 
 def test_bad_tolerance():
@@ -1333,9 +1361,6 @@ def test_bad_tolerance():
     with pytest.raises(ValueError):
         m.tol = -1
 
-    with pytest.raises(ValueError):
-        m.tol = 0
-
 
 def test_issue_544():
     m = Minuit(func0, x=0, y=0)
@@ -1479,3 +1504,76 @@ def test_minos_without_migrad():
     assert me.is_valid
     assert me.lower == approx(-2, abs=5e-3)
     assert me.upper == approx(2, abs=5e-3)
+
+
+def test_missing_ndata():
+    def fcn(a):
+        return a
+
+    fcn.errordef = 1
+
+    m = Minuit(fcn, 1)
+    assert_equal(m.ndof, np.nan)
+
+
+def test_issue_648():
+    class F:
+        errordef = 1
+        first = True
+
+        def __call__(self, a, b):
+            if self.first:
+                assert a == 1.0 and b == 2.0
+                self.first = False
+            return a ** 2 + b ** 2
+
+    m = Minuit(F(), a=1, b=2)
+    m.fixed["a"] = False  # this used to change a to b
+    m.migrad()
+
+
+def test_call_limit_reached_in_hesse():
+    m = Minuit(lambda x: ((x - 1.2) ** 4).sum(), np.ones(10) * 10)
+    m.errordef = 1
+    m.migrad(ncall=200)
+    assert m.fmin.has_reached_call_limit
+    assert m.fmin.nfcn < 205
+
+
+def test_issue_643():
+    def fcn(x, y, z):
+        return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2
+
+    fcn.errordef = Minuit.LEAST_SQUARES
+
+    m = Minuit(fcn, x=2, y=3, z=4)
+    m.migrad()
+
+    m2 = Minuit(fcn, x=m.values["x"], y=m.values["y"], z=m.values["z"])
+    # used to call MnHesse when it was not needed and quickly exhaust call limit
+    for i in range(10):
+        m2.minos()
+
+    m2.reset()
+    # used to exhaust call limit, because calls to MnHesse did not reset call count
+    for i in range(10):
+        m2.values = m.values
+        m2.minos()
+
+
+def test_issue_669():
+    m = Minuit(func0, x=0, y=0)
+
+    m.migrad()
+
+    xy1 = m.mncontour(x="x", y="y", size=10)
+    xy2 = m.mncontour(x="y", y="x", size=10)  # used to fail
+
+    # needs better way to compare polygons
+    for x, y in xy1:
+        match = False
+        for (y2, x2) in xy2:
+            if abs(x - x2) < 1e-3 and abs(y - y2) < 1e-3:
+                match = True
+                break
+        assert match
